/****************************************************************************
** libmatroska : parse Matroska files, see http://www.matroska.org/
**
** <file/class description>
**
** Copyright (C) 2002-2003 Steve Lhomme.  All rights reserved.
**
** This file may be distributed under the terms of the Q Public License
** as defined by Trolltech AS of Norway and appearing in the file
** LICENSE.QPL included in the packaging of this file.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** Licensees holding an other license may use this file in accordance with 
** the Agreement provided with the Software.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.matroska.org/license/qpl/ for QPL licensing information.
** See http://www.matroska.org/license/gpl/ for GPL licensing information.
**
** Contact license@matroska.org if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/
/*!
	\file
	\version \$Id: mkxds.cpp,v 1.4 2003/05/03 21:17:23 robux4 Exp $
	\author Steve Lhomme     <robux4 @ users.sf.net>
*/

#include <streams.h>
#include <initguid.h>    // declares DEFINE_GUID to declare an EXTERN_C const.
                         // Use once per project.
#include <tchar.h>
#include "mkxds.h"

#include "WinIOCallback.h"
#include "MatroskaReader.h"
#include "mkx_opin.h"

// ----------------------------------------------------------------------------
// Data used to register the filter
// ----------------------------------------------------------------------------

// Object table - all com objects in this DLL
CFactoryTemplate g_Templates[]=
{
	{ L"Matroska Demuxer Filter"
		, &CLSID_MKXDEMUX
		, MkxFilter::CreateInstance
		, NULL
		, &MkxFilter::sudFilter 
	}
};
int g_cTemplates = sizeof(g_Templates)/sizeof(g_Templates[0]);

const AMOVIESETUP_MEDIATYPE sudOutPinTypes = {
	&MEDIATYPE_NULL,
	&MEDIASUBTYPE_NULL
};

const AMOVIESETUP_PIN sudPins =
{
    L"Output",              // Obsolete
    FALSE,                  // Is this pin rendered ?
    TRUE,                   // Is it an output pin ?
    FALSE,                  // Can the filter create zero instance ?
    FALSE,                  // Does the filter create multiple instances ?
    &CLSID_NULL,            // Obsolete
    NULL,                   // Obsolete
    1,                      // Number of types
    &sudOutPinTypes };      // Pin details

// setup data - allows the self-registration to work.
const AMOVIESETUP_FILTER MkxFilter::sudFilter = {
    g_Templates[0].m_ClsID		// clsID
  , g_Templates[0].m_Name		// strName
  , MERIT_NORMAL				// dwMerit
  , 1							// nPins
  , &sudPins					// lpPin
};

// The streams.h DLL entrypoint.
extern "C" BOOL WINAPI DllEntryPoint(HINSTANCE, ULONG, LPVOID);

// The entrypoint required by the MSVC runtimes. This is used instead
// of DllEntryPoint directly to ensure global C++ classes get initialised.
BOOL WINAPI DllMain(HANDLE hDllHandle, DWORD dwReason, LPVOID lpreserved) {
	
    return DllEntryPoint(reinterpret_cast<HINSTANCE>(hDllHandle), dwReason, lpreserved);
}

STDAPI DllRegisterServer()
{
	HKEY key;
	DWORD disp;

    // TODO : use StringFromGUID2 on CLSID instead of hardcoded GUID string

	// Identification by extension
	// HKEY_CLASSES_ROOT\Media Type\Extensions\.ext
	if (ERROR_SUCCESS == RegCreateKeyEx(HKEY_CLASSES_ROOT,
			_T("Media Type\\Extensions\\.mkv"), 0, _T("REG_SZ"),
			REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &key, &disp))
	{
		// We are the source filter
		const BYTE guidSource[] = "{34293064-02F2-41d5-9D75-CC5967ACA1AB}";
		RegSetValueExA(key, "Source Filter", 0, REG_SZ, guidSource, countof(guidSource));
		RegCloseKey(key);
	}

	if (ERROR_SUCCESS == RegCreateKeyEx(HKEY_CLASSES_ROOT,
		_T("Media Type\\Extensions\\.mka"), 0, _T("REG_SZ"),
		REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &key, &disp))
	{
		// We are the source filter
		const BYTE guidSource[] = "{34293064-02F2-41d5-9D75-CC5967ACA1AB}";
		RegSetValueExA(key, "Source Filter", 0, REG_SZ, guidSource, countof(guidSource));
		RegCloseKey(key);
	}

	return AMovieDllRegisterServer2(TRUE);
}

STDAPI DllUnregisterServer()
{
	RegDeleteKey(HKEY_CLASSES_ROOT, _T("Media Type\\Extensions\\.mkv"));
	RegDeleteKey(HKEY_CLASSES_ROOT, _T("Media Type\\Extensions\\.mka"));

    return AMovieDllRegisterServer2(FALSE);
}

// ***********************************************************
// 
//  The MkxDs methods
//

// Provide the way for the COM support routines in <streams.h>
// to create a CKaxDemuxFilter object
CUnknown * WINAPI MkxFilter::CreateInstance(LPUNKNOWN punk, HRESULT *phr)
{
	DbgSetModuleLevel(LOG_ERROR,5);
    DbgSetModuleLevel(LOG_TRACE,5); // comment this to remove trace
	//DbgSetModuleLevel(LOG_MEMORY,2);
	DbgSetModuleLevel(LOG_LOCKING,5);
	
    NOTE("MkxFilter::CreateInstance");

    CUnknown *pNewObject = new MkxFilter(NAME("MkxDs Object"), punk, phr );
    if (pNewObject == NULL) {
        *phr = E_OUTOFMEMORY;
    }

    return pNewObject;
}

MkxFilter::MkxFilter(TCHAR *tszName, LPUNKNOWN punk, HRESULT *phr)
 :CSource(tszName, punk, *sudFilter.clsID, phr)
 ,m_pMKReader(NULL)
 ,m_FileHandle(NULL)
{
	
}

// ============================================================================
// MkxFilter IFileSourceFilter interface
// ============================================================================

STDMETHODIMP MkxFilter::Load(LPCOLESTR lpwszFileName, const AM_MEDIA_TYPE *pmt)
{
	NOTE("MkxFilter::Load in...");
	CheckPointer(lpwszFileName, E_POINTER);	
	
	// lstrlenW is one of the few Unicode functions that works on win95
	int cch = lstrlenW(lpwszFileName) + 1;
	
#ifndef _UNICODE
	TCHAR *lpszFileName=0;
	lpszFileName = new char[cch * 2];
	if (!lpszFileName) {
		NOTE("MkxFilter::Load new lpszFileName failed E_OUTOFMEMORY");
		return E_OUTOFMEMORY;
	}
	WideCharToMultiByte(GetACP(), 0, lpwszFileName, -1,
		lpszFileName, cch, NULL, NULL);
	NOTE1("MkxFilter::Load Loading %s", lpszFileName);
#else
	TCHAR lpszFileName[MAX_PATH]={0};
	lstrcpy(lpszFileName, lpwszFileName);
#endif
	
	m_pFileName = new WCHAR[cch];
	
	if (m_pFileName!=NULL)
	{
		CopyMemory(m_pFileName, lpwszFileName, cch*sizeof(WCHAR));	
	} else {
		NOTE("CCAudioSource::Load filename is empty !");
		return VFW_E_CANNOT_RENDER;
	}

	// Open the file & create pin
	m_FileHandle = new WinIOCallback(m_pFileName, MODE_READ);

	if (m_FileHandle == NULL)
		return S_FALSE;

	m_pMKReader = new MatroskaReader;

	if (m_pMKReader == NULL)
		return S_FALSE;

	if (m_pMKReader->InitKaxFile(m_FileHandle) < 0)
		return S_FALSE;
	
	// create an output pin for each track
	CreateOutputPin();

	return S_OK;
}

// ----------------------------------------------------------------------------

STDMETHODIMP MkxFilter::GetCurFile(LPOLESTR * ppszFileName, AM_MEDIA_TYPE *pmt)
{
	NOTE("MkxFilter::GetCurFile");
	
	CheckPointer(ppszFileName, E_POINTER);
	*ppszFileName = NULL;
	
	if (m_pFileName!=NULL)
	{
		DWORD n = sizeof(WCHAR)*(1+lstrlenW(m_pFileName));
		
		*ppszFileName = (LPOLESTR) CoTaskMemAlloc( n );
		if (*ppszFileName!=NULL)
		{
			CopyMemory(*ppszFileName, m_pFileName, n);
		}
	}
	
	if (pmt!=NULL)
	{
		// TODO
		NOTE("MkxFilter::GetCurFile TODO");
	}
	
	return NOERROR;
}

// ============================================================================

HRESULT MkxFilter::CreateOutputPin()
{
	HRESULT	hr = S_OK;

//	m_rtDuration = m_pMKReader->GetDuration();

	for(int i=0; i < m_pMKReader->GetTrackNumber(); i++)
	{
		
		TCHAR szPinName[32];		
		KaxTrackInfoStruct* pTrackInfo = m_pMKReader->GetTrackInfo(i);
		
		// create a new outpin
		switch(pTrackInfo->Type)
		{
		case track_video:
			swprintf(szPinName, _T("Video %s"), pTrackInfo->Name.c_str()); break;
		case track_audio:
			swprintf(szPinName, _T("Audio %s"), pTrackInfo->Name.c_str()); break;
		case track_subtitle:
			swprintf(szPinName, _T("Subtitle %s"), pTrackInfo->Name.c_str()); break;
		}					
		
		//this, m_pLock, &hr, szPinName, pTrackInfo
		MkxOutPin *pPin = new MkxOutPin(szPinName, &hr, this, szPinName);
		
		if(FAILED(hr))
		{
			break;
		}

//		m_nPins++;
//		m_OutPins.AddTail(pPin);
		AddPin(pPin);
	}

	return hr;
}
